/* $Aero: util/diskmark/diskmark.c,v 1.1.1.1 2002/03/06 00:06:09 brooks Exp $
 *
 * Copyright 2001 The Aerospace Corporation.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions, and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  The name of The Aerospace Corporation may not be used to endorse or
 *     promote products derived from this software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AEROSPACE CORPORATION "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AEROSPACE CORPORATION BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <inttypes.h>
#include <limits.h>
#include <string.h>

#ifndef DM_DEF_MAGIC
#define DM_DEF_MAGIC "AERO"
#endif

#ifndef DM_DEF_PTYPE
#define DM_DEF_PTYPE 0xAD
#endif

#ifndef DM_DEF_PARTITION
#define DM_DEF_PARTITION 4
#endif

#ifndef DM_DEF_DISK
#define DM_DEF_DISK "/dev/ad0"
#endif

#define DM_MAGIC_LEN 4
#define DM_PART_ENTRY_LEN 16
#define DM_PART_ENTRY_BASE 446

struct partent {
	char		flags;
	char		chs_start[3];
	unsigned char	type;
	char		chs_end[3];
	uint32_t	lba_start;
	uint32_t	lba_stop;
};

void usage(void) __dead2;
int hex2str(const char *hex, char *str);
const char *str2hex(const char *str, size_t size, char *hex);
int xdigit2val(int c);
int val2xdigit(int x);
int isprintable(const char *str, size_t size);

int
main(int argc, char *argv[])
{
	char		ch;

	int		writemode;
	int		checkmode;
	char		magic[DM_MAGIC_LEN + 1];
	char		mbr[512];
	struct partent	*part_entry;
	unsigned char	ptype;
	int		partition;
	int16_t		major;
	uint16_t	minor;
	const char	*disk;
	long		tmp_long;
	unsigned long	tmp_ulong;
	size_t		part_offset;
	FILE		*fp;

	/* Set up the defaults. */
	writemode = 0;
	checkmode = 0;
	memcpy(magic, DM_DEF_MAGIC, DM_MAGIC_LEN);
	magic[DM_MAGIC_LEN] = '\0';
	ptype = DM_DEF_PTYPE;
	partition = DM_DEF_PARTITION;
	major = 0;
	minor = 0;

	/* Parse the command line. */
	while ((ch = getopt(argc, argv, "m:p:t:w")) != -1) {
		switch (ch) {
		case 'm':
			if (strlen(optarg) == DM_MAGIC_LEN) {
				memcpy(magic, optarg, DM_MAGIC_LEN);
			} else if (strlen(optarg) == 10 &&
			    optarg[0] == '0' && optarg[1] == 'x') {
				if (hex2str(optarg, magic) != DM_MAGIC_LEN)
					usage();
			} else
				usage();
			break;
		case 'p':
			tmp_long = strtol(optarg, NULL, 0);
			if (tmp_long < 1 || tmp_long > 4)
				usage();
			partition = tmp_long;
			break;
		case 't':
			tmp_long = strtol(optarg, NULL, 0);
			if (tmp_long < 0 || tmp_long > 255)
				usage();
			ptype = tmp_long;
			break;
		case 'w':
			writemode = 1;
			break;
		case '?':
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 2 || argc == 3) {
		if (!writemode)
			checkmode = 1;

		tmp_long = strtol(argv[0], NULL, 0);
		/* XXX: should bounds check. */
		argc--;
		argv++;

		tmp_ulong = strtoul(argv[0], NULL, 0);
		/* XXX: should bounds check. */
		argc--;
		argv++;

		major = (int16_t)tmp_long;
		minor = (int16_t)tmp_ulong;
	}

	switch (argc) {
	case 0:
		disk = DM_DEF_DISK;
		break;
	case 1:
		disk = argv[0];
		break;
	default:
		usage();
	}

	/* Read in the mbr */
	if ((fp = fopen(disk, "r")) == NULL) {
		perror(disk);
		usage();
	}
	if (fread(mbr, 512, 1, fp) != 1) {
		perror("error reading mbr");
		exit(1);
	}
	fclose(fp);

	part_offset = DM_PART_ENTRY_BASE + DM_PART_ENTRY_LEN*(partition - 1);
	part_entry = (struct partent*) (mbr + part_offset);

	if (writemode) {
		if (part_entry->type != ptype && part_entry->type != 0) {
			fprintf(stderr,
			    "%s partition %d is neither type 0 or type 0x%X\n",
			    disk, partition, ptype);
			exit(1);
		}

		/* Pack the partition entry */
		part_entry->flags = 0x00;
		memset(&part_entry->chs_start, 0,
		    sizeof(part_entry->chs_start));
		part_entry->type = ptype;
		memset(&part_entry->chs_end, 0,
		    sizeof(part_entry->chs_end));
		memcpy(magic, &part_entry->lba_start, DM_MAGIC_LEN);
		((int16_t*)&part_entry->lba_stop)[0] = major;
		((uint16_t*)&part_entry->lba_stop)[1] = minor;

		/* Write the modified mbr back */
		if ((fp = fopen(disk, "w+")) == NULL) {
			perror(disk);
			usage();
		}
		if (fwrite(mbr, 512, 1, fp) != 1) {
			perror("error reading mbr");
			exit(1);
		}
		fclose(fp);

		return (0);
	}

	if (part_entry->type != ptype) {
		fprintf(stderr,
		    "%s partition %d is the wrong type "
		    "(expected 0x%X, got 0x%X)\n",
		    disk, partition, part_entry->type, ptype);
		exit(1);
	}

	if (checkmode) {
		if (!memcmp(magic, &part_entry->lba_start, DM_MAGIC_LEN)) {
			fprintf(stderr, "diskmark: wrong magic number\n");
			exit(1);
		}
		if (major != ((int16_t*)&part_entry->lba_stop)[0]) {
			fprintf(stderr, "diskmark: wrong major\n");
			exit(1);
		}
		if (minor != ((uint16_t*)&part_entry->lba_stop)[1]) {
			fprintf(stderr, "diskmark: wrong minor\n");
			exit(1);
		}
	} else {
		memcpy(&part_entry->lba_start, magic, DM_MAGIC_LEN);
		major = ((int16_t*)&part_entry->lba_stop)[0];
		minor = ((uint16_t*)&part_entry->lba_stop)[1];

		printf("magic: %s\nmajor: %d\nminor: %d\n",
		    (isprintable(magic, DM_MAGIC_LEN) ?
		    magic : str2hex(magic, DM_MAGIC_LEN, NULL)), major, minor);
	}

	return (0);
}

void
usage(void)
{
	fprintf(stderr, "diskmark -w [-m magic] [-p partition] [-t type] "
	    "major minor [disk]\n");
	fprintf(stderr,
	    "diskmark [-p partition] [-t type] [major minor] [disk]\n");
	exit(1);
}

int
hex2str(const char *hex, char *str)
{
	size_t		count;

	if (hex[0] != '0' || hex[1] != 'x')
		return -1;

	hex += 2;

	count = 0;
	while (hex[0] != '\0' && hex[1] != '\0') {
		if (!isxdigit(hex[0]) || !isxdigit(hex[1]))
			return -1;

		str[count] = (xdigit2val(hex[0]) << 4) + xdigit2val(hex[1]);

		count++;
		hex += 2;
	}

	if (hex[0] != '\0')
		return -1;

	return count;
}

const char *
str2hex(const char *str, size_t size, char *hex)
{
	size_t		count;
	static size_t	lhsize;
	static char	*lh;
	char		*th;

	if (hex == NULL) {
		if (lhsize < size*2 + 3) {
			lhsize = size*2 + 3;
			lh = realloc(lh, lhsize);
			if (lh == NULL) {
				lhsize = 0;
				return NULL;
			}
		}
		th = lh;
	} else
		th = hex;

	th[0] = '0';
	th[1] = 'x';
	th += 2;

	for (count = 0; count < size; count++) {
		th[0] = val2xdigit((str[count] & 0xF0) >> 4);
		th[1] = val2xdigit(str[count] & 0x0F);
		th += 2;
	}

	th[0] = '\0';

	if (hex == NULL)
		return lh;
	else
		return hex;
}

int
xdigit2val(int c)
{
	c = tolower(c);

	switch (c) {
	case '0': return 0;
	case '1': return 1;
	case '2': return 2;
	case '3': return 3;
	case '4': return 4;
	case '5': return 5;
	case '6': return 6;
	case '7': return 7;
	case '8': return 8;
	case '9': return 9;
	case 'a': return 10;
	case 'b': return 11;
	case 'c': return 12;
	case 'd': return 13;
	case 'e': return 14;
	case 'f': return 15;
	default:  return -1;
	}
}

int
val2xdigit(int x)
{
	char xdigits[] = "0123456789ABCDEF";

	if (x < 0 || x > 15)
		return -1;

	return xdigits[x];
}

int
isprintable(const char *str, size_t size)
{
	size_t		count;

	for (count = 0; count < size; count++) {
		if (!isprint(str[count]))
			return 0;
	}

	/* Make sure the string is null terminated. */
	if (str[size] != '\0')
		return 0;

	return 1;
}
